from __future__ import annotations

from pathlib import Path
from tempfile import gettempdir
from typing import Any

import click
import pydantic

try:
    from typing import Literal      # type: ignore
except ImportError:
    from typing_extensions import Literal


_TEMP_DIR = Path(gettempdir())

_default_config_dir = lambda: click.get_app_dir('instawow')


class _Config(pydantic.BaseSettings):

    config_dir: Path
    addon_dir: Path
    temp_dir: Path = _TEMP_DIR / 'instawow'
    game_flavour: Literal['retail', 'classic']

    @pydantic.validator('config_dir', 'addon_dir')
    def __expand_paths(cls, value: Path) -> Path:
        return Path(value).expanduser().resolve()

    @classmethod
    def read(cls) -> _Config:
        dummy_config = cls(addon_dir='', game_flavour='retail')
        return cls.parse_raw(dummy_config.config_file.read_text(encoding='utf-8'))

    def __init__(__pydantic_self__, **values: Any) -> None:
        values = __pydantic_self__._build_values(values)
        if not values.get('config_dir'):
            values = {**values, 'config_dir': _default_config_dir()}
        super(pydantic.BaseSettings, __pydantic_self__).__init__(**values)

    def json(self, **kwargs: Any) -> str:
        kwargs = {'exclude': {'config_dir'}, 'indent': 2, **kwargs}
        return super().json(**kwargs)

    def write(self) -> _Config:
        for dir_ in (self.config_dir,
                     self.logger_dir,
                     self.plugin_dir,
                     self.temp_dir,):
            dir_.mkdir(exist_ok=True)

        self.config_file.write_text(self.json(), encoding='utf-8')
        return self

    @property
    def is_classic(self) -> bool:
        return self.game_flavour == 'classic'

    @property
    def config_file(self) -> Path:
        return self.config_dir / 'config.json'

    @property
    def logger_dir(self) -> Path:
        return self.config_dir / 'logs'

    @property
    def plugin_dir(self) -> Path:
        return self.config_dir / 'plugins'

    class Config:
        case_insensitive = True
        env_prefix = 'INSTAWOW_'


Config = _Config



from __future__ import annotations

from datetime import datetime
from enum import Enum
from functools import wraps
from itertools import takewhile
import json
from operator import itemgetter
import re
from typing import TYPE_CHECKING, cast
from typing import Any, Callable, ClassVar, Dict, List, NamedTuple, Optional, Set, Tuple

from loguru import logger
from yarl import URL

from . import exceptions as E
from .models import Pkg, PkgOptions
from .utils import ManagerAttrAccessMixin, gather, run_in_thread as t, slugify, bbegone, is_not_stale

if TYPE_CHECKING:
    from .manager import Manager


class Strategies(str, Enum):

    default = 'default'
    latest = 'latest file available'
    curse_latest_beta = 'latest beta quality file available from CurseForge'
    curse_latest_alpha = 'latest alpha quality file available from CurseForge'


class Defn(NamedTuple):

    source: str
    name: str
    strategy: Strategies = Strategies.default

    def __str__(self) -> str:
        return f'{self.source}:{self.name}'


def validate_strategy(method: Callable) -> Callable:
    @wraps(method)
    async def wrapper(self, defn: Defn, *args: Any, **kwargs: Any) -> Pkg:
        if defn.strategy in self.strategies:
            return await method(self, defn, *args, **kwargs)
        raise E.PkgStrategyUnsupported(defn.strategy)

    return wrapper


class _FileCacheMixin:

    async def _cache_json_response(self: Any, url: str, *args: Any) -> Any:
        from hashlib import md5

        filename = md5(url.encode()).hexdigest()
        path = self.config.temp_dir / f'{filename}.json'

        if await t(is_not_stale)(path, *args):
            text = await t(path.read_text)(encoding='utf-8')
        else:
            async with self.web_client.get(url, raise_for_status=True) as response:
                text = await response.text()
            await t(path.write_text)(text, encoding='utf-8')
        return json.loads(text)


class Resolver(ManagerAttrAccessMixin):

    source: ClassVar[str]
    name: ClassVar[str]
    strategies: ClassVar[Set[Strategies]]

    def __init__(self, *, manager: Manager) -> None:
        self.manager = manager

    @classmethod
    def decompose_url(cls, value: str) -> Optional[Tuple[str, str]]:
        raise NotImplementedError

    async def synchronise(self) -> Resolver:
        return self

    async def resolve(self, defns: List[Defn]) -> Dict[Defn, Any]:
        raise NotImplementedError


class CurseResolver(Resolver, _FileCacheMixin):

    source = 'curse'
    name = 'CurseForge'
    strategies = {Strategies.default, Strategies.latest,
                  Strategies.curse_latest_beta,
                  Strategies.curse_latest_alpha}

    # https://twitchappapi.docs.apiary.io/
    addon_api_url = 'https://addons-ecs.forgesvc.net/api/v2/addon'
    slugs_url = ('https://raw.githubusercontent.com/layday/instascrape/data/'
                 'curseforge-slugs.json')   # v1

    _slugs: Optional[Dict[str, str]] = None

    @classmethod
    def decompose_url(cls, value: str) -> Optional[Tuple[str, str]]:
        url = URL(value)
        if (url.host == 'www.wowace.com'
                and len(url.parts) > 2
                and url.parts[1] == 'projects'):
            return (cls.source, url.parts[2].lower())
        elif (url.host == 'www.curseforge.com'
                and len(url.parts) > 3
                and url.parts[1:3] == ('wow', 'addons')):
            return (cls.source, url.parts[3].lower())

    async def synchronise(self) -> CurseResolver:
        if self._slugs is None:
            slugs = await self._cache_json_response(self.slugs_url, 8, 'hours')
            self._slugs = {k: str(v) for k, v in slugs.items()}
        return self

    async def resolve(self, defns: List[Defn]) -> Dict[Defn, Any]:
        slugs = cast(dict, self._slugs)
        ids_from_defns = {d: slugs.get(d.name, d.name) for d in defns}
        numeric_ids = list({i for i in ids_from_defns.values() if i.isdigit()})
        async with self.web_client.post(self.addon_api_url, json=numeric_ids) as response:
            if response.status == 404:
                json_response = []       # type: ignore
            else:
                json_response = await response.json()

        api_results = {str(r['id']): r for r in json_response}
        coros = (self.resolve_one(d, api_results.get(i))
                 for d, i in ids_from_defns.items())
        results = dict(zip(defns, await gather(coros)))
        return results

    @validate_strategy
    async def resolve_one(self, defn: Defn, metadata: Optional[dict]) -> Pkg:
        if not metadata:
            raise E.PkgNonexistent
        elif not metadata['latestFiles']:
            raise E.PkgFileUnavailable('no files available for download')

        def is_not_libless(file: dict) -> bool:
            # There's also an 'isAlternate' field that's missing some
            # 50 lib-less files from c. 2008.  'exposeAsAlternative' is
            # absent from the /file endpoint
            return not file['exposeAsAlternative']

        classic_version_prefix = '1.13'
        flavor = 'wow_classic' if self.config.is_classic else 'wow_retail'

        def is_compatible_with_game_version(file: dict) -> bool:
            # Files can belong both to retail and classic
            # but ``gameVersionFlavor`` can only be one of
            # 'wow_retail' or 'wow_classic'.  To spice things up,
            # ``gameVersion`` might not be populated so we still have to check
            # the value of ``gameVersionFlavor``
            return (file['gameVersionFlavor'] == flavor
                    or any(v.startswith(classic_version_prefix) is self.config.is_classic
                           for v in file['gameVersion']))

        # 1 = stable; 2 = beta; 3 = alpha
        if defn.strategy is Strategies.default:
            def has_release_type(f: dict) -> bool: return f['releaseType'] == 1
        elif defn.strategy is Strategies.curse_latest_beta:
            def has_release_type(f: dict) -> bool: return f['releaseType'] == 2
        elif defn.strategy is Strategies.curse_latest_alpha:
            def has_release_type(f: dict) -> bool: return f['releaseType'] == 3
        else:
            def has_release_type(f: dict) -> bool: return True

        files = (f for f in metadata['latestFiles']
                 if is_not_libless(f)
                 and is_compatible_with_game_version(f)
                 and has_release_type(f))
        try:
            # The ``id`` is just a counter so we don't have to go digging around dates
            file = max(files, key=itemgetter('id'))
        except ValueError:
            raise E.PkgFileUnavailable(f'no files compatible with {self.config.game_flavour} '
                                       f'using {defn.strategy.name!r} strategy')

        return Pkg(origin=self.source,
                   id=metadata['id'],
                   slug=metadata['slug'],
                   name=metadata['name'],
                   description=metadata['summary'],
                   url=metadata['websiteUrl'],
                   file_id=file['id'],
                   download_url=file['downloadUrl'],
                   date_published=file['fileDate'],
                   version=file['displayName'],
                   options=PkgOptions(strategy=defn.strategy.name))


class WowiResolver(Resolver, _FileCacheMixin):

    source = 'wowi'
    name = 'WoWInterface'
    strategies = {Strategies.default}

    # https://api.mmoui.com/v3/globalconfig.json
    list_api_url = 'https://api.mmoui.com/v3/game/WOW/filelist.json'
    details_api_url = URL('https://api.mmoui.com/v3/game/WOW/filedetails/')

    _files: Optional[Dict[str, dict]] = None

    @classmethod
    def decompose_url(cls, value: str) -> Optional[Tuple[str, str]]:
        url = URL(value)
        if (url.host in {'wowinterface.com', 'www.wowinterface.com'}
                and len(url.parts) == 3
                and url.parts[1] == 'downloads'):
            if url.name == 'landing.php':
                id_ = url.query.get('fileid')
                if id_:
                    return (cls.source, id_)
            elif url.name == 'fileinfo.php':
                id_ = url.query.get('id')
                if id_:
                    return (cls.source, id_)
            else:
                match = re.match(r'^(?:download|info)(?P<id>\d+)(?:-(?P<name>[^\.]+))?',
                                 url.name)
                if match:
                    id_, slug = match.groups()
                    if slug:
                        id_ = slugify(f'{id_} {slug}')
                    return (cls.source, id_)

    async def synchronise(self) -> WowiResolver:
        if self._files is None:
            files = await self._cache_json_response(self.list_api_url, 3600)
            self._files = {i['UID']: i for i in files}
        return self

    async def resolve(self, defns: List[Defn]) -> Dict[Defn, Any]:
        files = cast(dict, self._files)
        ids_from_defns = {d: ''.join(takewhile(str.isdigit, d.name)) for d in defns}
        numeric_ids = {i for i in ids_from_defns.values() if i.isdigit()}
        url = self.details_api_url / f'{",".join(numeric_ids)}.json'
        async with self.web_client.get(url) as response:
            if response.status == 404:
                json_response = []       # type: ignore
            else:
                json_response = await response.json()

        api_results = {r['UID']: {**files[r['UID']], **r} for r in json_response}
        coros = (self.resolve_one(d, api_results.get(i))
                 for d, i in ids_from_defns.items())
        results = dict(zip(defns, await gather(coros)))
        return results

    @validate_strategy
    async def resolve_one(self, defn: Defn, metadata: Optional[dict]) -> Pkg:
        if not metadata:
            raise E.PkgNonexistent
        if metadata['UIPending'] == '1':
            raise E.PkgFileUnavailable('new file awaiting approval')

        # The file is only assumed to be compatible with classic if 'WoW Classic'
        # is listed under compatibility, and incompatible with retail if
        # 'WoW Classic' is the sole element in the array.  I'm not sure
        # when 'UICompatibility' was added but it's very often not populated
        # for retail add-ons *shrugsies*
        compatibility = {e['name'] for e in metadata['UICompatibility'] or ()}
        if self.config.is_classic:
            if 'WoW Classic' not in compatibility:
                raise E.PkgFileUnavailable('file is not compatible with classic')
        elif compatibility and not compatibility - {'WoW Classic'}:
            raise E.PkgFileUnavailable('file is only compatible with classic')

        return Pkg(origin=self.source,
                   id=metadata['UID'],
                   slug=slugify(f'{metadata["UID"]} {metadata["UIName"]}'),
                   name=metadata['UIName'],
                   description=bbegone(metadata['UIDescription']),
                   url=metadata['UIFileInfoURL'],
                   file_id=metadata['UIMD5'],
                   download_url=metadata['UIDownload'],
                   date_published=metadata['UIDate'],
                   version=metadata['UIVersion'],
                   options=PkgOptions(strategy=defn.strategy.name))


class TukuiResolver(Resolver):

    source = 'tukui'
    name = 'Tukui'
    strategies = {Strategies.default}

    api_url = URL('https://www.tukui.org/api.php')

    @classmethod
    def decompose_url(cls, value: str) -> Optional[Tuple[str, str]]:
        url = URL(value)
        if (url.host == 'www.tukui.org'
                and url.path in {'/addons.php', '/classic-addons.php', '/download.php'}):
            id_or_slug = url.query.get('id') or url.query.get('ui')
            if id_or_slug:
                return (cls.source, id_or_slug)

    async def resolve(self, defns: List[Defn]) -> Dict[Defn, Any]:
        results = await gather(self.resolve_one(d) for d in defns)
        return dict(zip(defns, results))

    @validate_strategy
    async def resolve_one(self, defn: Defn) -> Pkg:
        id_, *_ = defn.name.partition('-')
        is_ui = id_ in {'elvui', 'tukui'}

        if self.config.is_classic:
            query = 'classic-addon'
        else:
            if is_ui:
                query = 'ui'
            else:
                query = 'addon'
        url = self.api_url.with_query({query: id_})
        async with self.web_client.get(url) as response:
            if not response.content_length:
                raise E.PkgNonexistent
            addon = await response.json(content_type='text/html')

        if is_ui:
            slug = id_
            date_published = datetime.fromisoformat(addon['lastupdate'])
        else:
            slug = slugify(f'{addon["id"]} {addon["name"]}')
            date_published = addon['lastupdate']

        return Pkg(origin=self.source,
                   id=addon['id'],
                   slug=slug,
                   name=addon['name'],
                   description=addon['small_desc'],
                   url=addon['web_url'],
                   file_id=addon['lastupdate'],
                   download_url=addon['url'],
                   date_published=date_published,
                   version=addon['version'],
                   options=PkgOptions(strategy=defn.strategy.name))


class InstawowResolver(Resolver):

    source = 'instawow'
    name = 'instawow'
    strategies = {Strategies.default}

    _addons = {('0', 'weakauras-companion'),
               ('1', 'weakauras-companion-autoupdate')}

    @classmethod
    def decompose_url(cls, value: str) -> None:
        return

    async def resolve(self, defns: List[Defn]) -> Dict[Defn, Any]:
        results = await gather(self.resolve_one(d) for d in defns)
        return dict(zip(defns, results))

    @validate_strategy
    async def resolve_one(self, defn: Defn) -> Pkg:
        try:
            id_, slug = next(p for p in self._addons if defn.name in p)
        except StopIteration:
            raise E.PkgNonexistent

        from .wa_updater import WaCompanionBuilder

        builder = WaCompanionBuilder(self.manager)
        if id_ == '1':
            try:
                await builder.build()
            except ValueError as error:
                raise E.PkgFileUnavailable('account name not provided') from error

        return Pkg(origin=self.source,
                   id=id_,
                   slug=slug,
                   name='WeakAuras Companion',
                   description='A WeakAuras Companion clone.',
                   url='https://github.com/layday/instawow',
                   file_id=await t(builder.checksum)(),
                   download_url=builder.file_out.as_uri(),
                   date_published=datetime.now(),
                   version='1.0.0',
                   options=PkgOptions(strategy=defn.strategy.name))



from __future__ import annotations

import asyncio
from pathlib import Path
from typing import TYPE_CHECKING
from typing import Any, Dict, List, NamedTuple, Optional, Sequence

from loguru import logger
from pydantic import BaseModel, BaseSettings, Extra, validator
from yarl import URL

from .utils import ManagerAttrAccessMixin, bucketise, run_in_thread as t

if TYPE_CHECKING:
    from .manager import Manager


metadata_api = URL('https://data.wago.io/api/check/weakauras')
raw_api = URL('https://data.wago.io/api/raw/encoded')


class BuilderConfig(BaseSettings):

    account: str

    class Config:
        case_insensitive = True
        env_prefix = 'WAC_'

    def _build_values(self, init_kwargs: Dict[str, Any]) -> Dict[str, Any]:
        return {**init_kwargs, **self._build_environ()}     # Prioritise env vars


class AuraEntry(BaseModel):

    id: str
    uid: str
    parent: Optional[str]
    url: URL
    version: int
    semver: Optional[str]
    ignore_wago_update: bool = False

    class Config:
        arbitrary_types_allowed = True
        fields = {'ignore_wago_update' : {'alias': 'ignoreWagoUpdate'}}

    @validator('url', pre=True)
    def __convert_url(cls, value: str) -> URL:
        return URL(value)

    @classmethod
    def from_lua_ast(cls, values: dict) -> Optional[AuraEntry]:
        return cls(**values) if values.get('url') else None


class ApiMetadata__Changelog(BaseModel):

    format: Optional[str]
    text: str = ''

    class Config:
        extra = Extra.forbid


class ApiMetadata(BaseModel):

    id: str
    name: str
    slug: str
    url: str
    created: str
    modified: str
    game: str
    fork_of: Optional[str]
    username: str
    version: int
    version_string: str
    changelog: ApiMetadata__Changelog
    region_type: Optional[str]

    class Config:
        extra = Extra.forbid
        fields = {'id': {'alias': '_id'},
                  'fork_of': {'alias': 'forkOf'},
                  'version_string': {'alias': 'versionString'},
                  'region_type': {'alias': 'regionType'}}


class _OutdatedAura(NamedTuple):

    slug: str
    existing_auras: List[AuraEntry]
    metadata: ApiMetadata
    import_string: str


class WaCompanionBuilder(ManagerAttrAccessMixin):
    """A WeakAuras Companion port for shellfolk."""

    def __init__(self, manager: Manager) -> None:
        self.manager = manager
        self.builder_dir = self.manager.config.plugin_dir / __name__
        self.file_out = self.builder_dir / 'WeakAurasCompanion.zip'

    @staticmethod
    def extract_auras(source: str) -> Dict[str, List[AuraEntry]]:
        from lupa import LuaRuntime

        lua_runtime = LuaRuntime()
        table = lua_runtime.eval(source[source.find('= ') + 1 :])
        raw_auras = map(dict, table['displays'].values())
        auras = filter(None, map(AuraEntry.from_lua_ast, raw_auras))
        aura_groups = {k: v
                       for k, v in bucketise(auras, key=lambda a: a.url.parts[1]).items()
                       if not any(i.ignore_wago_update for i in v)}
        return aura_groups

    def extract_installed_auras(self, account: str) -> Dict[str, List[AuraEntry]]:
        import time

        accounts = self.config.addon_dir.parents[1] / 'WTF/Account'
        saved_vars = accounts / account / 'SavedVariables/WeakAuras.lua'

        start = time.perf_counter()
        aura_groups = self.extract_auras(saved_vars.read_text(encoding='utf-8'))
        logger.debug(f'auras extracted in {time.perf_counter() - start}s')
        return aura_groups

    async def get_wago_aura_metadata(self, aura_ids: Sequence[str]) -> List[ApiMetadata]:
        url = metadata_api.with_query({'ids': ','.join(aura_ids)})
        async with self.web_client.get(url) as response:
            metadata = await response.json()

        results = dict.fromkeys(aura_ids)
        for item in (ApiMetadata(**i) for i in metadata):
            if item.slug in results:
                results[item.slug] = item
            else:
                logger.info(f'extraneous {item.slug!r} slug in metadata')
        return list(results.values())

    async def get_wago_aura_import_string(self, aura_id: str) -> str:
        url = raw_api.with_query({'id': aura_id})
        async with self.web_client.get(url) as response:
            return await response.text()

    async def get_outdated(self, aura_groups: Dict[str, List[AuraEntry]]) -> List[_OutdatedAura]:
        if not aura_groups:
            return []

        aura_metadata = await self.get_wago_aura_metadata(list(aura_groups))
        auras_with_metadata = filter(lambda v: v[1],
                                     zip(aura_groups.items(), aura_metadata))
        outdated_auras = [((s, w), r)
                          for (s, w), r in auras_with_metadata
                          if r.version > next((a for a in w if not a.parent), w[0]).version]
        new_auras = await asyncio.gather(*(self.get_wago_aura_import_string(r.id)
                                           for _, r in outdated_auras))
        return [_OutdatedAura(s, w, r, p)
                for ((s, w), r), p in zip(outdated_auras, new_auras)]

    def make_addon(self, outdated: List[_OutdatedAura]) -> None:
        from jinja2 import Environment, FileSystemLoader
        from zipfile import ZipFile, ZipInfo

        tpl_dir = Path(__file__).parent / 'wa_templates'
        jinja_env = Environment(loader=FileSystemLoader(str(tpl_dir)),
                                trim_blocks=True, lstrip_blocks=True)

        with ZipFile(self.file_out, 'w') as addon_zip:
            def write_tpl(filename: str, ctx: dict) -> None:
                # Not using a plain string as the first argument to ``writestr``
                # 'cause the timestamp is set dynamically by default, which
                # renders the build rather - shall we say - unreproducible
                zip_info = ZipInfo(filename=f'WeakAurasCompanion/{filename}')
                tpl = jinja_env.get_template(filename)
                addon_zip.writestr(zip_info, tpl.render(**ctx))

            write_tpl('data.lua',
                      {'was': [(o.metadata.slug,
                                {'name': o.metadata.name,
                                 'author': o.metadata.username,
                                 'encoded': o.import_string,
                                 'wagoVersion': o.metadata.version,
                                 'wagoSemver': o.metadata.version_string,
                                 'versionNote': o.metadata.changelog.text})
                               for o in outdated],
                       'uids': [(a.uid, a.url.parts[1])
                                for o in outdated
                                for a in o.existing_auras],
                       'ids':  [(a.id, a.url.parts[1])
                                for o in outdated
                                for a in o.existing_auras],
                       'stash': []})    # Send to WAC not supported - always empty
            write_tpl('init.lua', {})
            write_tpl('WeakAurasCompanion.toc',
                      {'interface': '11302' if self.config.is_classic else '80205'})

    async def build(self, account: Optional[str] = None) -> None:
        config = BuilderConfig(account=account)
        auras = await t(self.extract_installed_auras)(config.account)
        outdated_auras = await self.get_outdated(auras)

        await t(self.builder_dir.mkdir)(exist_ok=True)
        await t(self.make_addon)(outdated_auras)

    def checksum(self) -> str:
        from hashlib import sha256
        return sha256(self.file_out.read_bytes()).hexdigest()



from __future__ import annotations

import asyncio
from collections import defaultdict
from datetime import datetime, timedelta
from functools import partial, reduce
from itertools import repeat
from pathlib import Path
import re
from typing import TYPE_CHECKING
from typing import (Any, Awaitable, Callable, Iterable, List, NamedTuple,
                    Optional, Tuple, Type, TypeVar, Union)

try:
    from typing import Literal      # type: ignore
except ImportError:
    from typing_extensions import Literal

if TYPE_CHECKING:
    import prompt_toolkit.shortcuts.progress_bar.base as pbb

    from .manager import CliManager


class ManagerAttrAccessMixin:

    def __getattr__(self, name: str) -> Any:
        return getattr(self.manager, name)


class _TocEntry(NamedTuple):

    key: str
    value: Optional[str]

    def __bool__(self) -> bool:
        return bool(self.value)


class TocReader:
    """Extracts key–value pairs from TOC files."""

    def __init__(self, contents: str, default: Literal[None, ''] = '') -> None:
        possible_entries = (map(str.strip, e.lstrip('#').partition(':')[::2])
                            for e in contents.splitlines()
                            if e.startswith('##'))
        self.entries = {k: v for k, v in possible_entries if k}
        self.default = default

    def __getitem__(self, key: Union[str, Tuple[str, ...]]) -> _TocEntry:
        if isinstance(key, tuple):
            try:
                return next(filter(lambda i: i.value, (self[k] for k in key)))
            except StopIteration:
                key = key[0]
        return _TocEntry(key, self.entries.get(key, self.default))

    @classmethod
    def from_path(cls, path: Path, *args: Any, **kwargs: Any) -> TocReader:
        return cls(path.read_text(encoding='utf-8-sig', errors='replace'),
                   *args, **kwargs)

    @classmethod
    def from_path_name(cls, path: Path, *args: Any, **kwargs: Any) -> TocReader:
        return cls.from_path(path / f'{path.name}.toc', *args, **kwargs)


O = TypeVar('O')


class cached_property:

    def __init__(self, f: Callable) -> None:
        self.f = f

    def __get__(self, o: Optional[O], t: Type[O]) -> Any:
        if o is None:
            return self.f
        else:
            o.__dict__[self.f.__name__] = v = self.f(o)
            return v


def bucketise(iterable: Iterable, key: Callable = (lambda v: v)) -> defaultdict:
    "Place the elements of ``iterable`` into a bucket according to ``key``."
    bucket = defaultdict(list)      # type: ignore
    for value in iterable:
        bucket[key(value)].append(value)
    return bucket


def dict_merge(*args: dict) -> dict:
    "Right merge any number of ``dict``s."
    return reduce(lambda p, n: {**p, **n}, args, {})


async def gather(it: Iterable, return_exceptions: bool = True) -> List[Any]:
    return await asyncio.gather(*it, return_exceptions=return_exceptions)


def run_in_thread(fn: Callable) -> Callable[..., Awaitable]:
    return lambda *a, **k: asyncio.get_running_loop().run_in_executor(None, partial(fn, *a, **k))


_match_loweralphanum = re.compile(r'[^0-9a-z ]')


def slugify(text: str) -> str:
    "Convert an add-on name into a lower-alphanumeric slug."
    return '-'.join(_match_loweralphanum.sub(' ', text.casefold()).split())


_match_bbcode = re.compile(r'''
(\[(?:(?:font
        |size
        |color
        |list
        |url
        |email
        |highlight)=[^\]]*
     |(?:b
        |i
        |u
        |left
        |center
        |right
        |list
        |\*
        |url
        |quote
        |code))\]
 |\[/(?:font
       |size
       |color
       |b
       |i
       |u
       |left
       |center
       |right
       |list
       |list
       |url
       |email
       |quote
       |code
       |highlight)\]
|\[img\][^\[]*\[/img\])
''', re.IGNORECASE | re.VERBOSE)


def bbegone(text: str) -> str:
    "Naïvely remove BBCode from package descriptions."
    return _match_bbcode.sub('', text).strip()


def make_progress_bar(**kwargs: Any) -> pbb.ProgressBar:
    "A ``ProgressBar`` with download progress expressed in megabytes."
    from contextlib import contextmanager
    from prompt_toolkit.formatted_text import HTML
    import prompt_toolkit.shortcuts.progress_bar.base as pbb
    from prompt_toolkit.shortcuts.progress_bar import formatters

    # There is a race condition in the bar's shutdown logic
    # where the bar-drawing thread is left to run indefinitely
    # if the prompt_toolkit app is not (yet) running
    # but is scheduled to run when ``__exit__``ing the bar.
    # If an exception occurs early in the execution cycle or if execution
    # finishes before prompt_toolkit is able to crank the app,
    # instawow will hang.  This is my ham-fisted attempt to work
    # around all that by relocating the ``.exit()`` call.  The progress bar
    # 'signals' to the thread that the app is ready to exit
    # and the app is exited from inside the thread.
    # To avoid having to monkey-patch ``run`` which is nested
    # inside ``__enter__``, we overwrite ``pbb._auto_refresh_context``.

    class ProgressBar(pbb.ProgressBar):
        def __exit__(self, *args):
            if self._has_sigwinch:
                self._loop.add_signal_handler(pbb.signal.SIGWINCH, self._previous_winch_handler)

            # Signal to ``_auto_refresh_context`` that it should exit the app
            self.app._should_exit = True
            self._thread.join()
            del self.app._should_exit

    _bar_refresh_interval = .1

    @contextmanager
    def _auto_refresh_context(app, _refresh_interval=None):
        done = [False]

        def run():
            while not done[0]:
                pbb.time.sleep(_bar_refresh_interval)
                app.invalidate()
                if getattr(app, '_should_exit', False) and app.is_running:
                    app.exit()

        t = pbb.threading.Thread(target=run)
        t.daemon = True
        t.start()

        try:
            yield
        finally:
            done[0] = True

    pbb._auto_refresh_context = _auto_refresh_context

    class ProgressInMB(formatters.Progress):
        template = f'{formatters.Progress.template}MB'

        def format(self, _bar, counter, _width):
            html = HTML(self.template)
            return html.format(**{f: f'{getattr(counter, f) / 2 ** 20:.1f}'
                                  for f in ('current', 'total')})

    formatters = [formatters.Label(),
                  formatters.Text(' '),
                  formatters.Percentage(),
                  formatters.Text(' '),
                  formatters.Bar(),
                  formatters.Text(' '),
                  ProgressInMB(),
                  formatters.Text(' '),
                  formatters.Text('eta [', style='class:time-left'),
                  formatters.TimeLeft(),
                  formatters.Text(']', style='class:time-left'),
                  formatters.Text(' '),]
    progress_bar = ProgressBar(formatters=formatters, **kwargs)
    return progress_bar


def is_not_stale(path: Path, ttl: int, unit: str = 'seconds') -> bool:
    "Check if a file is older than ``ttl``."
    mtime = path.exists() and path.stat().st_mtime
    return mtime > 0 and (datetime.now() - datetime.fromtimestamp(mtime)) < timedelta(**{unit: ttl})


def is_outdated(manager: CliManager) -> bool:
    """Check against PyPI to see if `instawow` is outdated.

    The response is cached for 24 hours.
    """
    from . import __version__

    def parse_version(version: str) -> Tuple[int, ...]:
        return tuple(map(int, version.split('.')[:3]))

    cache_file = manager.config.config_dir / '.pypi_version'
    if is_not_stale(cache_file, 1, 'days'):
        version = cache_file.read_text(encoding='utf-8')
    else:
        from aiohttp.client import ClientError

        async def get_metadata() -> dict:
            url = 'https://pypi.org/pypi/instawow/json'
            async with manager.web_client.get(url) as response:
                return await response.json()

        try:
            version = manager.run(get_metadata())['info']['version']
        except ClientError:
            version = __version__
        else:
            cache_file.write_text(version, encoding='utf-8')

    # Installed version > version on PyPI if running in dev or
    # user upgraded without being prompted
    if parse_version(__version__) > parse_version(version):
        return False
    else:
        return __version__ != version


def setup_logging(logger_dir: Path, level: Union[int, str] = 'INFO') -> int:
    from loguru import logger

    handler = {'sink': logger_dir / 'error.log',
               'level': level,
               'rotation': '1 MB',
               'enqueue': True}
    handler_id, = logger.configure(handlers=(handler,))
    return handler_id



from __future__ import annotations

from enum import Enum
from functools import partial, update_wrapper
from itertools import chain, islice
from operator import itemgetter
from textwrap import fill
from typing import TYPE_CHECKING
from typing import (Any, Callable, Generator, Iterable, List, NamedTuple,
                    Optional, Sequence, Set, Tuple, Union)

import click

from . import __version__
from .config import Config
from . import exceptions as E
from .manager import CliManager, prepare_db_session
from .models import Pkg, PkgFolder, is_pkg
from .resolvers import Strategies, Defn
from .utils import TocReader, bucketise, cached_property, is_outdated, setup_logging


class Symbols(str, Enum):

    SUCCESS = click.style('✓', fg='green')
    FAILURE = click.style('✗', fg='red')
    WARNING = click.style('!', fg='blue')

    @classmethod
    def from_result(cls, result: E.ManagerResult) -> Symbols:
        if isinstance(result, E.InternalError):
            return cls.WARNING
        elif isinstance(result, E.ManagerError):
            return cls.FAILURE
        return cls.SUCCESS

    def __str__(self) -> str:
        return self.value


class Report:

    def __init__(self, results: Sequence[Tuple[str, E.ManagerResult]],
                 filter_fn: Callable = (lambda _: True)) -> None:
        self.results = results
        self.filter_fn = filter_fn

    @property
    def code(self) -> int:
        return any(r for _, r in self.results
                   if (isinstance(r, (E.ManagerError, E.InternalError))
                       and self.filter_fn(r)))

    def __str__(self) -> str:
        return '\n'.join(
            (f'{Symbols.from_result(r)} {click.style(str(a), bold=True)}\n'
             f'  {r.message}')
            for a, r in self.results
            if self.filter_fn(r)
            )

    def generate(self) -> None:
        report = str(self)
        if report:
            click.echo(report)

    def generate_and_exit(self) -> None:
        self.generate()
        ctx = click.get_current_context()
        ctx.exit(self.code)


def tabulate(rows: Iterable, *, show_index: bool = True) -> str:
    from texttable import Texttable, Texttable as c

    table = Texttable(max_width=0).set_deco(c.BORDER | c.HEADER | c.VLINES)
    if show_index:
        iter_rows = iter(rows)
        header = next(iter_rows)
        rows = [('', *header), *((i, *v) for i, v in enumerate(iter_rows, start=1))]
        table.set_cols_align(('r', *('l' for _ in header)))
    return table.add_rows(rows).draw()


def decompose_pkg_defn(ctx: click.Context, param: Any,
                       value: Union[str, List[str], Tuple[str, ...]], *,
                       raise_when_invalid: bool = True) -> Any:
    if isinstance(value, (list, tuple)):
        return list(bucketise(decompose_pkg_defn(ctx, param, v) for v in value))    # Remove dupes

    if ':' not in value:
        if raise_when_invalid:
            raise click.BadParameter(value)

        parts = ('*', value)
    else:
        for resolver in ctx.obj.m.resolvers.values():
            parts = resolver.decompose_url(value)
            if parts:
                break
        else:
            parts = value.partition(':')[::2]
    return Defn(*parts)


def decompose_pkg_defn_with_strat(ctx, param, value):
    defns = decompose_pkg_defn(ctx, param, [d for _, d in value])
    strategies = (Strategies[s] for s, _ in value)
    return [Defn(o, n, s) for (o, n, _), s in zip(defns, strategies)]


def get_pkg_from_substr(manager: CliManager, defn: Defn) -> Optional[Pkg]:
    pkg = manager.get(defn)
    pkg = pkg or (manager.db_session.query(Pkg).filter(Pkg.slug.contains(defn.name))
                  .order_by(Pkg.name).first())
    return pkg


def _pass_manager(f: Callable) -> Callable:
    def new_func(*args: Any, **kwargs: Any) -> Callable:
        return f(click.get_current_context().obj.m, *args, **kwargs)
    return update_wrapper(new_func, f)


class _AddonType(str):
    "Dummy class used to customise the help text in Click."

_AddonType.__name__ = 'ADDON'


@click.group(context_settings={'help_option_names': ('-h', '--help')})
@click.version_option(__version__, prog_name='instawow')
@click.option('--debug',
              is_flag=True, default=False,
              help='Log more things.')
@click.pass_context
def main(ctx, debug):
    "Add-on manager for World of Warcraft."
    if not ctx.obj:
        @object.__new__
        class ManagerSingleton:
            @cached_property
            def manager(self) -> CliManager:
                while True:
                    try:
                        config = Config.read().write()
                    except FileNotFoundError:
                        ctx.invoke(write_config)
                    else:
                        break

                setup_logging(config.logger_dir, 'DEBUG' if debug else 'INFO')
                db_session = prepare_db_session(config)
                manager = CliManager(config, db_session)
                if is_outdated(manager):
                    click.echo(f'{Symbols.WARNING} instawow is out of date')
                return manager

            m = manager

        ctx.obj = ManagerSingleton


@main.command()
@click.option('--strategy', '-s', 'strategic_addons',
              multiple=True,
              type=(click.Choice({s.name for s in Strategies}), _AddonType),
              callback=decompose_pkg_defn_with_strat,
              help=('One of '
                    + ', '.join(f"'{s.name}'" for s in islice(Strategies, 1, None))
                    + '; and followed by an add-on definition.'))
@click.option('--replace', '-o',
              is_flag=True, default=False,
              help='Replace existing add-ons.')
@click.argument('addons',
                nargs=-1, callback=decompose_pkg_defn)
@_pass_manager
def install(manager, addons, strategic_addons, replace) -> None:
    "Install add-ons."
    deduped_addons = bucketise(chain(strategic_addons, addons), key=itemgetter(0, 1))
    values = [v for v, *_ in deduped_addons.values()]
    results = manager.install(values, replace)
    Report(results.items()).generate_and_exit()


@main.command()
@click.argument('addons',
                nargs=-1, callback=decompose_pkg_defn)
@_pass_manager
def update(manager, addons) -> None:
    "Update installed add-ons."
    if addons:
        values = addons
    else:
        values = [Defn(p.origin, p.slug) for p in manager.db_session.query(Pkg).all()]
    results = manager.update(values)
    # Hide package from output if up to date and ``update`` was invoked without args
    filter_fn = lambda r: addons or not isinstance(r, E.PkgUpToDate)
    Report(results.items(), filter_fn).generate_and_exit()


@main.command()
@click.argument('addons',
                nargs=-1, required=True, callback=decompose_pkg_defn)
@_pass_manager
def remove(manager, addons) -> None:
    "Uninstall add-ons."
    results = manager.remove(addons)
    Report(results.items()).generate_and_exit()


@main.command()
@click.pass_context
def reconcile(ctx) -> None:
    "Reconcile add-ons."
    from asyncio import gather
    from functools import reduce
    from itertools import chain, repeat, starmap
    from .prompts import Choice, confirm, select, skip

    class _Addon(NamedTuple):
        name: str
        reader: TocReader

        @property
        def version(self) -> str:
            return self.reader['Version', 'X-Packaged-Version'].value

    manager = ctx.obj.m
    resolve = partial(manager.resolve, strategy=Strategies.default)
    install = partial(manager.install, strategy=Strategies.default, replace=True)
    TocReader_ = lambda n: TocReader.from_path_name(manager.config.addon_dir / n)

    def _prompt(addons: Sequence[_Addon], pkgs: Sequence[Pkg]) -> Union[Tuple[()], Tuple[Sequence[_Addon], Defn]]:
        def create_choice(pkg):
            defn = Defn(pkg.origin, pkg.slug)
            title = [('', str(defn)),
                     ('', '=='),
                     ('class:hilite' if highlight_version else '', pkg.version),]
            return Choice(title, (addons, defn), pkg=pkg)

        # Highlight version if there's multiple of them
        highlight_version = len(bucketise(i.version for i in chain(addons, pkgs))) > 1
        choices = list(chain(map(create_choice, pkgs), (skip,)))
        addon = addons[0]
        # Use 'unsafe_ask' to let ^C bubble up
        selection = select(f'{addon.name} [{addon.version or "?"}]', choices).unsafe_ask()
        return selection

    def prompt(groups: Iterable[Tuple[Sequence[_Addon], Sequence[Pkg]]]) -> Generator[Tuple[Sequence[_Addon], Defn], None, None]:
        for addons, results in groups:
            shortlist = [r for r in results if is_pkg(r)]
            if shortlist:
                selection = _prompt(addons, shortlist)
                selection and (yield selection)

    def match_toc_ids_from_sources(leftovers: Set[str]) -> Iterable[Tuple[List[_Addon], Any]]:
        "Attempt to match add-ons from host IDs contained in TOC files."
        ids_to_sources = {'X-WoWI-ID': 'wowi',
                          'X-Tukui-ProjectID': 'tukui',
                          'X-Curse-Project-ID': 'curse',}

        def merge_ids_and_dirs(buckets, match):
            dirs, ids = match
            for index, (old_dirs, old_ids) in enumerate(buckets):
                if old_ids & ids:
                    buckets[index] = (old_dirs + dirs, old_ids | ids)
                    break
            else:
                buckets.append(match)
            return buckets

        dir_tocs = ((n, TocReader_(n)) for n in sorted(leftovers))
        maybe_ids = (((n, r), (r[i] for i in ids_to_sources)) for n, r in dir_tocs)
        buckets = reduce(merge_ids_and_dirs,
                         (([t], {Defn(ids_to_sources[v.key], v.value) for v in i if v})
                          for t, i in maybe_ids),
                         [])
        results = manager.run(resolve(set(chain.from_iterable(i for _, i in buckets))))
        groups = ((list(starmap(_Addon, k)), list(map(results.__getitem__, v)))
                  for k, v in buckets)
        return groups

    def match_dir_names(leftovers: Set[str]) -> Iterable[Tuple[List[_Addon], Any]]:
        "Attempt to match folders against the CurseForge and WoWInterface catalogue."
        async def fetch_catalogues(*urls):
            async def fetch(url):
                async with manager.web_client.get(url) as response:
                    # The CurseForge catalogue has a content type of 'text/plain' -
                    # aiohttp's default behaviour is to raise when it's not application/json
                    return await response.json(content_type=None)

            return await gather(*map(fetch, urls))

        def merge_dirs_and_ids(buckets, match):
            dirs, id_ = match
            for index, (old_dirs, old_ids) in enumerate(buckets):
                if old_dirs & dirs:
                    buckets[index] = (old_dirs | dirs, old_ids | {id_})
                    break
            else:
                buckets.append((dirs, {id_}))
            return buckets

        urls = (manager.resolvers['curse'].folders_url,
                manager.resolvers['wowi'].list_api_url,)
        curse_catalogue, wowi_catalogue = manager.run(fetch_catalogues(*urls))
        dirs = chain(((set(e['UIDir']), Defn('wowi', e['UID']))
                      for e in wowi_catalogue),
                     ((set(i[2]), Defn('curse', str(i[0])))
                      for i in curse_catalogue
                      if manager.config.game_flavour in i[1]))
        buckets = reduce(merge_dirs_and_ids,
                         ((d & leftovers, u) for d, u in dirs if d & leftovers),
                         [])
        results = manager.run(resolve(list({u for _, i in buckets for u in i})))
        groups = ((sorted(_Addon(n, TocReader_(n)) for n in k),
                   [results[i] for i in v])
                  for k, v in buckets)
        return groups

    def get_leftovers() -> Set[str]:
        addons = {f.name for f in manager.config.addon_dir.iterdir() if f.is_dir()}
        leftovers = addons - {f.name for f in manager.db_session.query(PkgFolder).all()}
        return leftovers

    def match_all():
        # Match in order of increasing heuristicitivenessitude
        for fn in (match_toc_ids_from_sources,
                   match_dir_names,):
            groups = fn(get_leftovers())
            yield [u for _, u in prompt(groups)]

    if not get_leftovers():
        click.echo('No add-ons left to reconcile.')
        return

    click.echo('''\
- Use the arrow keys to navigate, <o> to open an
  add-on in your browser and enter to make a selection.
- Versions that differ from the installed version
  or differ between choices are highlighted in purple.
- instawow will do a first pass of all of your add-ons
  looking for source IDs in TOC files, e.g. X-Curse-Project-ID.
- If it is unable to reconcile all of your add-ons
  it will perform a second pass to match add-on folders
  against the CurseForge and WoWInterface catalogues.
- Selected add-ons will be reinstalled.
- This feature is experimental and very much untested.\
''')
    for selections in match_all():
        if selections and confirm('Install selected add-ons?').unsafe_ask():
            results = install(selections)
            Report(results.items()).generate()
    click.echo('- Unreconciled add-ons can be listed with '
                '`instawow list-folders -e`.')


@main.command()
@click.option('--limit', '-l',
              default=5, type=click.IntRange(1, 20, clamp=True),
              help='A number to limit results to.')
@click.argument('search-terms',
                nargs=-1, required=True, callback=lambda _, __, v: ' '.join(v))
@click.pass_context
def search(ctx, limit, search_terms):
    "Search for add-ons."
    from .prompts import Choice, checkbox, confirm

    manager = ctx.obj.m

    pkgs = manager.run(manager.search(search_terms, limit))
    if pkgs:
        choices = [Choice(f'{p.name}  ({d}=={p.version})', d, pkg=p)
                   for d, p in pkgs.items()]
        selections = checkbox('Select add-ons to install', choices=choices).unsafe_ask()
        if selections and confirm('Install selected add-ons?').unsafe_ask():
            ctx.invoke(install, addons=selections)


@main.command('list')
@click.option('--column', '-c', 'columns',
              multiple=True,
              help='A field to show in a column.  Nested fields are '
                   'dot-delimited.  Repeatable.')
@click.option('--columns', '-C', 'print_columns',
              is_flag=True, default=False,
              help='Print a list of all possible column values.')
@click.option('--sort-by', '-s', 'sort_key',
              default='name',
              help='A key to sort the table by.  '
                   'You can chain multiple keys by separating them with a comma '
                   'just as you would in SQL, '
                   'e.g. `--sort-by="origin, date_published DESC"`.')
@_pass_manager
def list_installed(manager, columns, print_columns, sort_key) -> None:
    "List installed add-ons."
    from operator import attrgetter
    from sqlalchemy import inspect, text

    def format_columns(pkg):
        for column in columns:
            try:
                value = attrgetter(column)(pkg)
            except AttributeError:
                raise click.BadParameter(column, param_hint=['--column', '-c'])
            if column == 'folders':
                yield '\n'.join(f.name for f in value)
            elif column == 'options':
                yield f'strategy = {value.strategy}'
            elif column == 'description':
                yield fill(value, width=50, max_lines=3)
            else:
                yield value

    if print_columns:
        columns = [('field',),
                   *((c,) for c in (*inspect(Pkg).columns.keys(),
                                    *inspect(Pkg).relationships.keys()))]
        click.echo(tabulate(columns, show_index=False))
    else:
        pkgs = manager.db_session.query(Pkg).order_by(text(sort_key)).all()
        if pkgs:
            rows = [('add-on', *columns),
                    *((Defn(p.origin, p.slug), *format_columns(p)) for p in pkgs)]
            click.echo(tabulate(rows))


@main.command()
@click.option('--exclude-own', '-e',
              is_flag=True, default=False,
              help='Exclude folders managed by instawow.')
@click.option('--toc-entry', '-t', 'toc_entries',
              multiple=True,
              help='An entry to extract from the TOC.  Repeatable.')
@_pass_manager
def list_folders(manager, exclude_own, toc_entries) -> None:
    "List add-on folders."
    folders = {f for f in manager.config.addon_dir.iterdir() if f.is_dir()}
    if exclude_own:
        folders -= {manager.config.addon_dir / f.name
                    for f in manager.db_session.query(PkgFolder).all()}

    folder_tocs = ((n, n / f'{n.name}.toc') for n in folders)
    folder_readers = sorted((n, TocReader.from_path(t)) for n, t in folder_tocs if t.exists())
    if folder_readers:
        rows = [('folder',
                 *(f'[{e}]' for e in toc_entries)),
                *((n.name,
                  *(fill(t[e].value, width=50) for e in toc_entries))
                  for n, t in folder_readers)]
        click.echo(tabulate(rows))


@main.command()
@click.option('--toc-entry', '-t', 'toc_entries',
              multiple=True,
              help='An entry to extract from the TOC.  Repeatable.')
@click.argument('addon', callback=partial(decompose_pkg_defn,
                                          raise_when_invalid=False))
@_pass_manager
def info(manager, addon, toc_entries) -> None:
    "Show detailed add-on information."
    pkg = get_pkg_from_substr(manager, addon)
    if pkg:
        rows = {'name': pkg.name,
                'source': pkg.origin,
                'id': pkg.id,
                'slug': pkg.slug,
                'description': fill(pkg.description, max_lines=5),
                'homepage': pkg.url,
                'version': pkg.version,
                'release date': pkg.date_published,
                'folders': '\n'.join(f.name for f in pkg.folders),
                'strategy': pkg.options.strategy}

        if toc_entries:
            for folder in pkg.folders:
                toc_reader = TocReader.from_path_name(manager.config.addon_dir / folder.name)
                rows.update({f'[{folder.name} {k}]': fill(toc_reader[k].value)
                             for k in toc_entries})
        click.echo(tabulate([(), *rows.items()], show_index=False))
    else:
        Report([(addon, E.PkgNotInstalled())]).generate_and_exit()


@main.command()
@click.argument('addon', callback=partial(decompose_pkg_defn,
                                          raise_when_invalid=False))
@_pass_manager
def visit(manager, addon) -> None:
    "Open an add-on's homepage in your browser."
    pkg = get_pkg_from_substr(manager, addon)
    if pkg:
        import webbrowser
        webbrowser.open(pkg.url)
    else:
        Report([(addon, E.PkgNotInstalled())]).generate_and_exit()


@main.command()
@click.argument('addon', callback=partial(decompose_pkg_defn,
                                          raise_when_invalid=False))
@_pass_manager
def reveal(manager, addon) -> None:
    "Open an add-on folder in your file manager."
    pkg = get_pkg_from_substr(manager, addon)
    if pkg:
        import webbrowser
        webbrowser.open((manager.config.addon_dir / pkg.folders[0].name).as_uri())
    else:
        Report([(addon, E.PkgNotInstalled())]).generate_and_exit()


@main.command()
def write_config() -> None:
    "Configure instawow."
    import os.path
    from prompt_toolkit.completion import PathCompleter, WordCompleter
    from prompt_toolkit.shortcuts import CompleteStyle, prompt
    from prompt_toolkit.validation import Validator

    prompt_ = partial(prompt, complete_style=CompleteStyle.READLINE_LIKE)

    class DirectoryCompleter(PathCompleter):
        def __init__(self, *args, **kwargs):
            super().__init__(*args,
                             expanduser=True, only_directories=True, **kwargs)

        def get_completions(self, document, complete_event):
            # Append slash to every completion
            for completion in super().get_completions(document, complete_event):
                completion.text += '/'
                yield completion

    def validate_addon_dir(value: str) -> bool:
        path = os.path.expanduser(value)
        return os.path.isdir(path) and os.access(path, os.W_OK)

    completer = DirectoryCompleter()
    validator = Validator.from_callable(validate_addon_dir,
                                        error_message='must be a writable directory')
    addon_dir = prompt_('Add-on directory: ',
                        completer=completer, validator=validator)

    game_flavours = ('retail', 'classic')
    completer = WordCompleter(game_flavours)
    validator = Validator.from_callable(game_flavours.__contains__,
                                        error_message='must be one of: retail, classic')
    game_flavour = prompt_('Game flavour: ',
                           completer=completer, validator=validator)

    config = Config(addon_dir=addon_dir, game_flavour=game_flavour).write()
    click.echo(f'Configuration written to: {config.config_file}')


@main.command()
@_pass_manager
def show_config(manager) -> None:
    "Show the active configuration."
    click.echo(manager.config.json(exclude=set()))


@main.group('weakauras-companion')
def _weakauras_group() -> None:
    "Manage your WeakAuras."


@_weakauras_group.command('build')
@click.option('--account', '-a',
              required=True,
              help='Your account name.  This is used to locate '
                   'the WeakAuras data file.')
@_pass_manager
def build_weakauras_companion(manager, account) -> None:
    "Build the WeakAuras Companion add-on."
    from .wa_updater import WaCompanionBuilder

    manager.run(WaCompanionBuilder(manager).build(account))


@_weakauras_group.command('list')
@click.option('--account', '-a',
              required=True,
              help='Your account name.  This is used to locate '
                   'the WeakAuras data file.')
@_pass_manager
def list_installed_wago_auras(manager, account) -> None:
    "List WeakAuras installed from Wago."
    from .wa_updater import WaCompanionBuilder

    aura_groups = WaCompanionBuilder(manager).extract_installed_auras(account)
    installed_auras = sorted((a.id, a.url, str(a.ignore_wago_update).lower())
                             for v in aura_groups.values()
                             for a in v
                             if not a.parent)
    click.echo(tabulate([('name', 'url', 'ignore updates'),
                         *installed_auras]))



from __future__ import annotations

from functools import partial
from typing import TYPE_CHECKING, Any, List, Optional

from prompt_toolkit.application import Application
from prompt_toolkit.key_binding import KeyBindings
from prompt_toolkit.keys import Keys
from prompt_toolkit.styles import Style
from questionary import Choice as _Choice, confirm as _confirm
from questionary.prompts.common import InquirerControl, Separator, create_inquirer_layout
from questionary.question import Question

if TYPE_CHECKING:
    from .models import Pkg


class Choice(_Choice):

    def __init__(self, *args: Any, pkg: Optional[Pkg] = None, **kwargs: Any) -> None:
        super().__init__(*args, **kwargs)
        self.pkg = pkg


qstyle = Style([('qmark', 'fg:ansicyan'),
                ('answer', 'fg: nobold'),
                ('hilite', 'fg:ansimagenta'),
                ('skipped', 'fg:ansiyellow'),
                ('question', 'nobold'),
                ('x-question', 'bold'),])
skip = Choice([('', 'skip')], ())

confirm = partial(_confirm, style=qstyle)


def checkbox(message: str, choices: List[Choice]) -> Question:
    # This is a cut-down version of ``questionary.checkbox`` with the addition
    # of an <o> key binding for opening package URLs

    def get_prompt_tokens():
        tokens = [('class:x-question', message),]
        if ic.is_answered:
            tokens.append(('class:answer', '  done'))
        else:
            tokens.append(('class:instruction',
                           '  (use arrow keys to move, '
                           '<space> to select, '
                           '<o> to view in your browser and '
                           '<i> to invert)'))
        return tokens

    ic = InquirerControl(choices, None,
                         use_indicator=False, use_shortcuts=False, use_pointer=True)
    bindings = KeyBindings()

    @bindings.add(Keys.ControlQ, eager=True)
    @bindings.add(Keys.ControlC, eager=True)
    def _(event):
        event.app.exit(exception=KeyboardInterrupt, style='class:aborting')

    @bindings.add(' ', eager=True)
    def toggle(event):
        pointed_choice = ic.get_pointed_at().value
        if pointed_choice in ic.selected_options:
            ic.selected_options.remove(pointed_choice)
        else:
            ic.selected_options.append(pointed_choice)

    @bindings.add('i', eager=True)
    def invert(event):
        inverted_selection = [c.value for c in ic.choices if
                              not isinstance(c, Separator)
                              and c.value not in ic.selected_options
                              and not c.disabled]
        ic.selected_options = inverted_selection

    @bindings.add(Keys.Down, eager=True)
    @bindings.add('j', eager=True)
    def move_cursor_down(event):
        ic.select_next()
        while not ic.is_selection_valid():
            ic.select_next()

    @bindings.add(Keys.Up, eager=True)
    @bindings.add('k', eager=True)
    def move_cursor_up(event):
        ic.select_previous()
        while not ic.is_selection_valid():
            ic.select_previous()

    @bindings.add(Keys.ControlM, eager=True)
    def set_answer(event):
        ic.is_answered = True
        event.app.exit(result=[c.value for c in ic.get_selected_values()])

    @bindings.add('o', eager=True)
    def open_url(event):
        pkg = ic.get_pointed_at().pkg
        if pkg:
            import webbrowser
            webbrowser.open(pkg.url)

    @bindings.add(Keys.Any)
    def other(event):
        # Disallow inserting other text
        pass

    layout = create_inquirer_layout(ic, get_prompt_tokens)
    app = Application(layout=layout, key_bindings=bindings, style=qstyle)
    return Question(app)


def select(message: str, choices: List[Choice]) -> Question:
    def get_prompt_tokens():
        tokens = [('', '- '),
                  ('class:x-question', message),]
        if ic.is_answered:
            answer = ''.join(t for _, t in ic.get_pointed_at().title)
            tokens += [('', '  '),
                       ('class:skipped' if answer == 'skip' else '', answer)]
        return tokens

    ic = InquirerControl(choices, None,
                         use_indicator=False, use_shortcuts=False, use_pointer=True)
    bindings = KeyBindings()

    @bindings.add(Keys.ControlQ, eager=True)
    @bindings.add(Keys.ControlC, eager=True)
    def _(event):
        event.app.exit(exception=KeyboardInterrupt, style='class:aborting')

    @bindings.add(Keys.Down, eager=True)
    @bindings.add('j', eager=True)
    def move_cursor_down(event):
        ic.select_next()
        while not ic.is_selection_valid():
            ic.select_next()

    @bindings.add(Keys.Up, eager=True)
    @bindings.add('k', eager=True)
    def move_cursor_up(event):
        ic.select_previous()
        while not ic.is_selection_valid():
            ic.select_previous()

    @bindings.add(Keys.ControlM, eager=True)
    def set_answer(event):
        ic.is_answered = True
        event.app.exit(result=ic.get_pointed_at().value)

    @bindings.add('o', eager=True)
    def open_url(event):
        pkg = ic.get_pointed_at().pkg
        if pkg:
            import webbrowser
            webbrowser.open(pkg.url)

    @bindings.add(Keys.Any)
    def other(event):
        # Disallow inserting other text
        pass

    layout = create_inquirer_layout(ic, get_prompt_tokens)
    app = Application(layout=layout, key_bindings=bindings, style=qstyle)
    return Question(app)



from alembic.command import (list_templates, init, revision, merge, upgrade,
                             downgrade, show, history, heads, branches, current,
                             stamp, edit)
from alembic.config import Config as AlembicConfig


def make_config(url: str) -> AlembicConfig:
    config = AlembicConfig()
    config.set_main_option('script_location', 'instawow:migrations')
    config.set_main_option('sqlalchemy.url', url)
    return config



from __future__ import annotations

from typing import TYPE_CHECKING, ClassVar, Sequence, Optional, Set

if TYPE_CHECKING:
    from .models import Pkg


class ManagerResult:

    fmt_message: ClassVar[str]

    @property
    def message(self) -> str:
        return self.fmt_message.format(self=self)


class PkgInstalled(ManagerResult):

    fmt_message = 'installed {self.new_pkg.version}'

    def __init__(self, new_pkg: Pkg) -> None:
        super().__init__()
        self.new_pkg = new_pkg


class PkgUpdated(ManagerResult):

    fmt_message = 'updated {self.old_pkg.version} to {self.new_pkg.version}'

    def __init__(self, old_pkg: Pkg, new_pkg: Pkg) -> None:
        super().__init__()
        self.old_pkg = old_pkg
        self.new_pkg = new_pkg


class PkgRemoved(ManagerResult):

    fmt_message = 'removed'

    def __init__(self, old_pkg: Pkg) -> None:
        super().__init__()
        self.old_pkg = old_pkg


class ManagerError(ManagerResult,
                   Exception):
    pass


class PkgAlreadyInstalled(ManagerError):

    fmt_message = 'package already installed'


class PkgConflictsWithInstalled(ManagerError):

    fmt_message = "package folders conflict with installed package's "\
                  '{self.conflicts[0].origin}:{self.conflicts[0].slug}'

    def __init__(self, conflicts: Sequence[Pkg]) -> None:
        super().__init__()
        self.conflicts = conflicts


class PkgConflictsWithUncontrolled(ManagerError):

    fmt_message = "package folders conflict with an add-on's"\
                  ' not controlled by instawow'

    def __init__(self, folders: Set[str]) -> None:
        super().__init__()
        self.folders = folders


class PkgNonexistent(ManagerError):

    fmt_message = 'package does not exist'


class PkgFileUnavailable(ManagerError):

    fmt_message = 'package file is not available for download'

    def __init__(self, detailed_message: Optional[str] = None) -> None:
        super().__init__()
        self.detailed_message = detailed_message

    @property
    def message(self) -> str:
        return self.detailed_message or super().message


class PkgNotInstalled(ManagerError):

    fmt_message = 'package is not installed'


class PkgOriginInvalid(ManagerError):

    fmt_message = 'package source is invalid'


class PkgUpToDate(ManagerError):

    fmt_message = 'package is up to date'


class PkgStrategyUnsupported(ManagerError):

    fmt_message = '{self.strategy.name!r} strategy is not valid for source'

    def __init__(self, strategy: str) -> None:
        super().__init__()
        self.strategy = strategy


class InternalError(ManagerResult,
                    Exception):

    fmt_message = 'instawow encountered an error'

    def __init__(self, error: BaseException) -> None:
        super().__init__()
        self.error = error



from __future__ import annotations

import asyncio
from contextlib import asynccontextmanager
import contextvars as cv
from functools import partial
from itertools import repeat
from pathlib import Path, PurePath
import posixpath
from shutil import move as _move
from tempfile import NamedTemporaryFile, mkdtemp
from typing import *

from loguru import logger
from sqlalchemy import create_engine
from sqlalchemy.orm import sessionmaker, scoped_session

from . import db_version
from . import exceptions as E
from .models import ModelBase, Pkg, PkgFolder, should_migrate, is_pkg
from .resolvers import (Defn, Strategies,
                        CurseResolver, WowiResolver, TukuiResolver, InstawowResolver)
from .utils import (bucketise, cached_property, dict_merge, gather,
                    make_progress_bar, run_in_thread)

if TYPE_CHECKING:
    from types import SimpleNamespace
    import aiohttp
    from .config import Config


USER_AGENT = 'instawow (https://github.com/layday/instawow)'

_web_client: cv.ContextVar[aiohttp.ClientSession] = cv.ContextVar('_web_client')

AsyncNamedTemporaryFile = partial(run_in_thread(NamedTemporaryFile), prefix='instawow-')
async_mkdtemp = run_in_thread(mkdtemp)
async_move = run_in_thread(_move)


@asynccontextmanager
async def open_temp_writer() -> AsyncGenerator[Tuple[Path, Callable], None]:
    fh = await AsyncNamedTemporaryFile(delete=False, suffix='.zip')
    try:
        yield (Path(fh.name), run_in_thread(fh.write))
    finally:
        await run_in_thread(fh.close)()


async def new_temp_dir(*args: Any, **kwargs: Any) -> PurePath:
    return PurePath(await async_mkdtemp(*args, **kwargs))


async def move(paths: Iterable[Path], dest: PurePath) -> None:
    for path in paths:
        await async_move(str(path), dest)


# macOS 'resource forks' are sometimes included in download zips - these
# violate our one add-on per folder contract-thing, and serve absolutely
# no purpose.  They won't be extracted and they won't be added to the database
_zip_excludes = {'__MACOSX'}


def _find_base_dirs(names):
    return {n for n in (posixpath.dirname(n) for n in names)
            if n and posixpath.sep not in n} - _zip_excludes


def _should_extract(base_dirs):
    def is_member(name):
        head, sep, _ = name.partition(posixpath.sep)
        return sep and head in base_dirs

    return is_member


def Archive(path: PurePath, delete_after: bool) -> Callable:
    from zipfile import ZipFile

    @asynccontextmanager
    async def enter() -> AsyncGenerator[Tuple[List[str], Callable], None]:
        zip_file = await run_in_thread(ZipFile)(path)
        names = zip_file.namelist()
        base_dirs = _find_base_dirs(names)

        def extract(parent: Path) -> None:
            conflicts = base_dirs & {f.name for f in parent.iterdir()}
            if conflicts:
                raise E.PkgConflictsWithUncontrolled(conflicts)
            else:
                members = filter(_should_extract(base_dirs), names)
                zip_file.extractall(parent, members=members)

        def exit_() -> None:
            zip_file.close()
            if delete_after:
                path.unlink()   # type: ignore

        try:
            yield (sorted(base_dirs), run_in_thread(extract))
        finally:
            await run_in_thread(exit_)()

    return enter


async def download_archive(pkg: Pkg, *, chunk_size: int = 4096) -> Callable:
    url = pkg.download_url
    if url.startswith('file://'):
        from urllib.parse import unquote

        path = PurePath(unquote(url[7:]))
        return Archive(path, delete_after=False)
    else:
        async with (_web_client.get()
                    .get(url, trace_request_ctx={'show_progress': True})) as response, \
                   open_temp_writer() as (path, write):
            async for chunk in response.content.iter_chunked(chunk_size):
                await write(chunk)
        return Archive(path, delete_after=True)


def prepare_db_session(config: Config) -> scoped_session:
    db_path = config.config_dir / 'db.sqlite'
    db_url = f'sqlite:///{db_path}'
    # We need to distinguish between newly-created databases and
    # databases predating alembic in instawow - attempting to migrate a
    # new database will throw an error
    db_exists = db_path.exists()

    engine = create_engine(db_url)
    ModelBase.metadata.create_all(engine)
    if should_migrate(engine, db_version):
        from .migrations import make_config, stamp, upgrade

        alembic_config = make_config(db_url)
        if db_exists:
            logger.info(f'migrating database to {db_version}')
            upgrade(alembic_config, db_version)
        else:
            logger.info(f'stamping database with {db_version}')
            stamp(alembic_config, db_version)

    return scoped_session(sessionmaker(bind=engine))


async def init_web_client(**kwargs: Any) -> aiohttp.ClientSession:
    from aiohttp import ClientSession, TCPConnector

    kwargs = {'connector': TCPConnector(force_close=True, limit_per_host=10),
              'headers': {'User-Agent': USER_AGENT},
              'trust_env': True,
              **kwargs}
    return ClientSession(**kwargs)


@object.__new__
class _DummyResolver:

    async def synchronise(self) -> _DummyResolver:
        return self

    async def resolve(self, defns: List[Defn], **kwargs: Any) -> Dict[Defn, E.PkgOriginInvalid]:
        return dict(zip(defns, repeat(E.PkgOriginInvalid())))


class _ResolverDict(dict):

    RESOLVERS = {CurseResolver, WowiResolver, TukuiResolver, InstawowResolver}

    def __init__(self, manager: Manager) -> None:
        super().__init__((r.source, r(manager=manager)) for r in self.RESOLVERS)

    def __missing__(self, key: Hashable) -> Type[_DummyResolver]:
        return _DummyResolver


def _error_wrapper(error: E.ManagerError) -> Callable:
    async def error_out() -> NoReturn:
        raise error
    return error_out


class Manager:

    def __init__(self, config: Config, db_session: scoped_session) -> None:
        self.config = config
        self.db_session = db_session
        self.resolvers = _ResolverDict(self)

    @property
    def web_client(self) -> aiohttp.ClientSession:
        return _web_client.get()

    @web_client.setter
    def web_client(self, value: aiohttp.ClientSession) -> None:
        _web_client.set(value)

    def get(self, defn: Defn) -> Pkg:
        "Retrieve a package from the database."
        return (self.db_session.query(Pkg)
                .filter(Pkg.origin == defn.source,
                        (Pkg.id == defn.name) | (Pkg.slug == defn.name))
                .first())

    async def resolve(self, defns: Sequence[Defn]) -> Dict[Defn, Any]:
        "Resolve definitions into packages."
        defns_by_source = bucketise(defns, key=lambda v: v.source)
        results = await gather([((await self.resolvers[s].synchronise())
                                 .resolve(list(b)))
                                for s, b in defns_by_source.items()])
        results_by_defn = dict_merge(dict.fromkeys(defns), *results)
        return results_by_defn

    async def search(self, search_terms: str, limit: int, *,
                     scorer: str = 'partial_ratio') -> Dict[Defn, Pkg]:
        "Search the combined names catalogue for packages."
        from fuzzywuzzy import fuzz, process
        from .resolvers import _FileCacheMixin as cache

        url = ('https://raw.githubusercontent.com/layday/instascrape/data/'
               'combined-names.json')   # v1
        combined_names = await cache._cache_json_response(self, url, 8, 'hours')
        defns_for_names = bucketise(((n, Defn(*map(str, d))) for n, d, f in combined_names
                                     if self.config.game_flavour in f),
                                    key=lambda v: v[0])

        matches = process.extract(search_terms, defns_for_names.keys(),
                                  limit=limit, scorer=getattr(fuzz, scorer))
        defns = [d
                 for m, _ in matches
                 for _, d in defns_for_names[m]]
        results = await self.resolve(defns)
        pkgs_by_defn = {Defn(r.origin, r.slug): r
                        for r in results.values() if is_pkg(r)}
        return pkgs_by_defn

    async def _install(self, pkg: Pkg, open_archive: Callable, replace: bool) -> E.PkgInstalled:
        async with open_archive() as (folders, extract):
            conflicts = (self.db_session.query(PkgFolder)
                         .filter(PkgFolder.name.in_(folders)).all())
            if conflicts:
                raise E.PkgConflictsWithInstalled([c.pkg for c in conflicts])
            if replace:
                temp_dir = await new_temp_dir(dir=self.config.temp_dir, prefix=f'{folders[0]}-')
                await move((self.config.addon_dir / f for f in folders), temp_dir)
            await extract(self.config.addon_dir)

        pkg.folders = [PkgFolder(name=f) for f in folders]
        self.db_session.add(pkg)
        self.db_session.commit()
        return E.PkgInstalled(pkg)

    async def prep_install(self, defns: Sequence[Defn], replace: bool) -> Dict[Defn, Callable]:
        "Retrieve packages to install."
        results = await self.resolve([d for d in defns if not self.get(d)])
        installables = {(d, r): download_archive(r) for d, r in results.items()
                        if is_pkg(r)}
        archives = await gather(installables.values())

        coros = (dict.fromkeys(defns, _error_wrapper(E.PkgAlreadyInstalled())),
                 {d: _error_wrapper(r) for d, r in results.items()},
                 {d: (lambda p=p, a=a: self._install(p, a, replace))
                  for (d, p), a in zip(installables, archives)})
        return dict_merge(*coros)

    async def _update(self, old_pkg: Pkg, new_pkg: Pkg, open_archive: Callable) -> E.PkgUpdated:
        async with open_archive() as (folders, extract):
            filter_ = (PkgFolder.name.in_(folders),
                       PkgFolder.pkg_origin != new_pkg.origin,
                       PkgFolder.pkg_id != new_pkg.id)
            conflicts = self.db_session.query(PkgFolder).filter(*filter_).first()
            if conflicts:
                raise E.PkgConflictsWithInstalled(conflicts.pkg)

            await self._remove(old_pkg)
            await extract(self.config.addon_dir)

        new_pkg.folders = [PkgFolder(name=f) for f in folders]
        self.db_session.add(new_pkg)
        self.db_session.commit()
        return E.PkgUpdated(old_pkg, new_pkg)

    async def prep_update(self, defns: Sequence[Defn]) -> Dict[Defn, Callable]:
        "Retrieve packages to update."
        def rebuild_defn(defn: Defn, pkg: Optional[Pkg]) -> Defn:
            return Defn(*defn[:2], Strategies[pkg.options.strategy]) if pkg else defn

        checked_defns = {rebuild_defn(c, p): p
                         for c, p in ((d, self.get(d)) for d in defns)}
        candidates = {d: p for d, p in checked_defns.items() if p}
        results = await self.resolve(list(candidates.keys()))
        installables = {d: r for d, r in results.items() if is_pkg(r)}
        updatables = {(d, candidates[d], p): download_archive(p)
                      for d, p in installables.items()
                      if p.file_id != candidates[d].file_id}
        archives = await gather(updatables.values())

        coros = (dict.fromkeys(checked_defns, _error_wrapper(E.PkgNotInstalled())),
                 {d: _error_wrapper(r) for d, r in results.items()},
                 {d: _error_wrapper(E.PkgUpToDate()) for d in installables},
                 {d: (lambda o=o, p=p, a=a: self._update(o, p, a))
                  for (d, o, p), a in zip(updatables, archives)})
        return dict_merge(*coros)

    async def _remove(self, pkg: Pkg) -> E.PkgRemoved:
        temp_dir = await new_temp_dir(dir=self.config.temp_dir,
                                      prefix=f'{pkg.folders[0].name}-')
        await move((self.config.addon_dir / f.name for f in pkg.folders), temp_dir)

        self.db_session.delete(pkg)
        self.db_session.commit()
        return E.PkgRemoved(pkg)

    async def prep_remove(self, defns: Sequence[Defn]) -> Dict[Defn, Callable]:
        "Prepare packages to remove."
        coros = (dict.fromkeys(defns, _error_wrapper(E.PkgNotInstalled())),
                 {d: (lambda p=p: self._remove(p))
                  for d, p in ((d, self.get(d)) for d in defns) if p})
        return dict_merge(*coros)


_tick_interval = .1
_tickers: cv.ContextVar[Set[asyncio.Task]] = cv.ContextVar('_tickers', default=set())


@asynccontextmanager
async def cancel_tickers() -> AsyncGenerator[None, None]:
    try:
        yield
    finally:
        for ticker in _tickers.get():
            ticker.cancel()


async def init_cli_web_client(*, manager: CliManager) -> aiohttp.ClientSession:
    from cgi import parse_header
    from aiohttp import TraceConfig

    def extract_filename(params: aiohttp.TraceRequestEndParams) -> str:
        cd = params.response.headers.get('Content-Disposition', '')
        _, cd_params = parse_header(cd)
        filename = cd_params.get('filename') or params.response.url.name
        return filename

    async def do_on_request_end(session: aiohttp.ClientSession,
                                ctx: SimpleNamespace,
                                params: aiohttp.TraceRequestEndParams) -> None:
        if ctx.trace_request_ctx and ctx.trace_request_ctx.get('show_progress'):
            bar = manager.bar(label=f'Downloading {extract_filename(params)}',
                              total=params.response.content_length)

            async def ticker(bar=bar, params=params) -> None:
                try:
                    while not params.response.content._eof:
                        bar.current = params.response.content._cursor
                        await asyncio.sleep(_tick_interval)
                finally:
                    bar.progress_bar.counters.remove(bar)

            tickers = _tickers.get()
            tickers.add(asyncio.create_task(ticker()))

    trace_config = TraceConfig()
    trace_config.on_request_end.append(do_on_request_end)
    trace_config.freeze()
    return await init_web_client(trace_configs=[trace_config])


async def _intercept(fn: Callable) -> E.ManagerResult:
    try:
        return await fn()
    except E.ManagerError as error:
        return error
    except Exception as error:
        logger.exception('internal error')
        return E.InternalError(error)


class CliManager(Manager):

    def __init__(self, config: Config, db_session: scoped_session,
                 progress_bar_factory: Optional[Callable] = None) -> None:
        super().__init__(config, db_session)
        self.progress_bar_factory = progress_bar_factory or make_progress_bar
        self.loop = asyncio.new_event_loop()

    @logger.catch(reraise=True)
    def _process(self, prepper: Callable, uris: Sequence[Defn],
                 *args: Any, **kwargs: Any) -> Dict[Defn, E.ManagerResult]:
        async def do_process():
            async with cancel_tickers():
                coros = await prepper(list(uris), *args, **kwargs)
                return {d: await _intercept(c) for d, c in coros.items()}

        with self.bar:
            return self.run(do_process())

    def run(self, awaitable: Awaitable) -> Any:
        "Run ``awaitable`` inside an explicit context."
        async def do_run():
            async with (await init_cli_web_client(manager=self)) as self.web_client:
                return await awaitable

        runner = lambda: self.loop.run_until_complete(do_run())
        return cv.copy_context().run(runner)

    @cached_property
    def bar(self) -> Any:
        return self.progress_bar_factory()

    @property
    def install(self) -> Callable:
        return partial(self._process, self.prep_install)

    @property
    def update(self) -> Callable:
        return partial(self._process, self.prep_update)

    @property
    def remove(self) -> Callable:
        return partial(self._process, self.prep_remove)



from functools import partial
import os
from pathlib import Path
from setuptools import find_packages, setup

if os.environ.get('INSTAWOW_DEV'):
    import fastentrypoints


read_text = partial(Path.read_text, encoding='utf-8')


setup(name='instawow',
      use_scm_version={'write_to': 'instawow/_version.py',
                       'write_to_template': "__version__ = '{version}'\n",},
      description='A CLI for managing World of Warcraft add-ons.',
      url='http://github.com/layday/instawow',
      author='layday',
      author_email='layday@protonmail.com',
      license='GPL-3.0-or-later',
      long_description=read_text(Path('README.rst')),
      python_requires='~=3.7',
      packages=find_packages(),
      setup_requires='setuptools_scm',
      install_requires=read_text(Path('requirements.txt')),
      include_package_data=True,
      entry_points={'console_scripts': ['instawow = instawow.cli:main']},)
